package extensions;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;

import java.net.URLDecoder;
import java.net.Socket;

import java.util.Collections;
import java.util.Comparator;
import java.util.Set;
import java.util.Map;
import java.util.SortedMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import dareka.common.Logger;
import dareka.common.Pair;
import dareka.common.regex.JavaMatcher;
import dareka.common.regex.JavaPattern;
import dareka.extensions.Extension2;
import dareka.extensions.ExtensionManager;
import dareka.extensions.RequestFilter;
import dareka.processor.HttpRequestHeader;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.Processor;
import dareka.processor.impl.Cache;
import dareka.processor.impl.EasyRewriter;

/*
 * totalRenameT|[g
 */
public class totalRenameHelper implements Extension2,
		Processor, RequestFilter, Comparator<String> {
	
	public static final int REVISION = 100915;
	public static final String VER_STRING = "totalRenameHelper_"+REVISION+"(v0.8)";
	public static final String LOG_PREFIX = "TR_Helper";
	public static final String PROP_DEBUG = "totalRenameDebug";
	
	private static final String[] PROCESSOR_SUPPORTED_METHODS = new String[] { "GET" };
	private static final Pattern PROCESSOR_SUPPORTED_PATTERN = Pattern.compile(
			"^http://www\\.nicovideo\\.jp/cache/tr_helper/(\\w+)\\?(.+)$");
	
	private static Set<String> deleted;
	private static final Pair<File, Long> deletedFileInfo =
		new Pair<File, Long>(ExtUtil.cacheFolderFile("#deleted.txt"), 0L);
	private static SortedMap<String, String> rminfo;
	private static final Pair<File, Long> rminfoFileInfo =
		new Pair<File, Long>(ExtUtil.cacheFolderFile("#rminfo.txt"), 0L);
	private static volatile boolean rminfoSaveNeeded;
	
	private static final Charset FILE_CHARSET = getCharset("totalRenameCharset");
	private static final StringResource resultOK = ExtUtil.responseText("OK");
	private static final StringResource resultNG = ExtUtil.responseText("NG");
	
	private static ExtUtil util;
	
	// Extension2 interface
	public void registerExtensions(ExtensionManager mgr) {
		final int REQUIRED_REVISION = 100909;
		if (ExtUtil.REVISION < REQUIRED_REVISION) {
			Logger.warning(LOG_PREFIX +
					": ERROR!!! ExtUtil_"+REQUIRED_REVISION+" OR LATER IS REQUIRED.");
		} else {
			if (util == null) {
				util = new ExtUtil(this, LOG_PREFIX, PROP_DEBUG);
				util.loadFile(rminfoFileInfo, "loadRemoveInfo", FILE_CHARSET);
			}
			mgr.registerProcessor(this);
			mgr.registerRequestFilter(this);
		}
	}
	
	public String getVersionString() {
		return VER_STRING;
	}
	
	// Processor Interface
	public String[] getSupportedMethods() {
		return PROCESSOR_SUPPORTED_METHODS;
	}
	
	public Pattern getSupportedURLAsPattern() {
		return PROCESSOR_SUPPORTED_PATTERN;
	}
	
	public String getSupportedURLAsString() {
		return null;
	}
	
	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		Matcher m = PROCESSOR_SUPPORTED_PATTERN.matcher(requestHeader.getURI());
		if (m.matches()) {
			util.loadFile(rminfoFileInfo, "loadRemoveInfo", FILE_CHARSET);
			String method = m.group(1);
			String[] params = m.group(2).split("-", 2);
			String param1 = (params.length > 1) ? URLDecoder.decode(params[1], "UTF-8") : null;
			if (method.equals("get")) {
				return methodGetTitle(params[0], param1);
			} else if (method.equals("rm")) {
				return methodRemove(params[0]);
			} else if (method.equals("rm_noajax")) {
				methodRemove(params[0]);
				return StringResource.getRedirect("http://www.nicovideo.jp/");
			} else if (method.equals("set")) {
				return methodSetReason(params[0], param1);
			} else if (method.equals("clear")) {
				setRemoveInfo(params[0], null);
				return resultOK;
			} else if (method.equals("rminfo")) {
				return methodRemoveInfo(params[0], param1);
			} else {
				util.info(method.concat(" not supported."));
			}
		}
		return StringResource.getNotFound();
	}
	
	// RequestFilter interface
	public int onRequest(HttpRequestHeader requestHeader) throws IOException {
		String uri = requestHeader.getURI();
		// GetOwnerName
		if (uri.startsWith("http://www.nicovideo.jp/allegation/allegation/")) {
			requestHeader.setURI(uri.replace("nicovideo", "smilevideo"));
			requestHeader.setMessageHeader("Host", "www.smilevideo.jp");
			requestHeader.removeMessageHeader("Cookie");
		}
		return RequestFilter.OK;
	}
	
	// Comparetor Interface
	public int compare(String s1, String s2) {
		int n1 = 0, n2 = 0;
		try {
			n1 = Integer.parseInt(s1.substring(2));
		} catch (NumberFormatException e) {}
		try {
			n2 = Integer.parseInt(s2.substring(2));
		} catch (NumberFormatException e) {}
		return n1 - n2;
	}
	
	// ExtUtil callback
	public void onMinutes() {
		onShutdown();
	}
	
	public void onShutdown() {
		if (rminfoSaveNeeded) {
			rminfoSaveNeeded = false;
			util.saveFile(rminfoFileInfo, rminfoString(), true, FILE_CHARSET);
		}
	}
	
	private static Charset getCharset(String propertyName) {
		Charset charset = Charset.defaultCharset();
		String charsetName = System.getProperty(propertyName);
		if (charsetName != null && !charsetName.equals("")) {
			charset = Charset.forName(charsetName);
		}
		return charset;
	}
	
	/**
	 * 폜}bv擾
	 * 
	 * @return MapIuWFNgԂ(keysmid, value폜R)
	 */
	public static Map<String, String> getRemoveInfo() {
		return Collections.unmodifiableSortedMap(rminfo);
	}
	
	/**
	 * LbVԂ擾
	 *
	 * /cache/tr_helper/get?<smid>-encodeURI(curTitle)
	 *  OK{ |:}^Cg
	 *  NG
	 *  RM[ 폜R(=폜̃^Cg)]
	 */
	private StringResource methodGetTitle(String smid, String curTitle) {
		String result = checkRemove(smid, curTitle);
		if (!result.startsWith("RM")) {
			Cache cache = new Cache(smid);
			if (cache.exists()) {
				result = "OK" + (isProtected(cache) ? ":" : " ") + cache.getTitle();
			}
		}
		util.debug("methodGetTitle: %s %s", smid, result);
		return ExtUtil.responseText(result, "text/plain; charset=utf-8");
	}
	
	private boolean isProtected(Cache cache) {
		if (containsDeleted(cache.getId()))
			return true;
		String cacheProtect = System.getProperty("TR_cacheProtect", "");
		if (cacheProtect.length() > 0) {
			String path = Cache.getPathFromId(cache.getId());
			if (path != null) {
				path += "/" + cache.getCacheFileName();
				try {
					Matcher m = Pattern.compile(cacheProtect).matcher(path);
					return m.find();
				} catch (PatternSyntaxException e) {}
			}
		}
		return false;
	}
	
	/**
	 * LbV폜
	 *
	 * /cache/tr_helper/rm?<smid>
	 *  OK
	 *  NG
	 */
	private StringResource methodRemove(String smid) throws IOException {
		Cache cache = new Cache(smid);
		String reason = cache.getTitle();
		boolean removed = Cache.removeAll(stripLow(smid));
		if (removed) {
			util.info(smid.concat(" removed."));
			setRemoveInfo(smid, "REMOVE: ".concat(reason));
			removeCommentCache(smid);
			return resultOK;
		}
		return resultNG;
	}
	
	/**
	 * RemoveInfoɍ폜Ro^
	 *
	 * /cache/tr_helper/set?<smid>-encodeURI(reason)
	 *  OK
	 *  NG
	 */
	private StringResource methodSetReason(String smid, String reason) {
		if (reason != null) {
			setRemoveInfo(smid, reason);
			return resultOK;
		}
		return resultNG;
	}
	
	/**
	 * RemoveInfo
	 *
	 * /cache/tr_helper/rminfo?{flash}
	 *  flush
	 *   ݂RemoveInfot@CɏoċɂBoeԂB
	 *   t@CGfB^ŕҏWOɎsƗǂ(폜ҏW̎flushȂƕ)
	 */
	private StringResource methodRemoveInfo(String mode, String param) {
		if (mode.equals("flush")) {
			String response = rminfoString();
			util.saveFile(rminfoFileInfo, response, true, FILE_CHARSET);
			return ExtUtil.responseText(response, "text/plain; charset=utf-8");
		}
		return StringResource.getNotFound();
	}
	
	// 폜Ǝ폜
	private String checkRemove(String smid, String title) {
		// ܂폜Ă΍폜Ȃ
		if (containsDeleted(smid)) {
			return "NG";
		}
		String result = "NG";
		// ɍ폜
		if (rminfo.containsKey(smid)) {
			result = "RM ".concat(getRemoveInfo(smid));
		}
		// Ōɖ
		if (result.equals("NG") && title != null) {
			JavaPattern[] p1 = EasyRewriter.getMatch("totalRenameHelperConf");
			JavaPattern[] p2 = EasyRewriter.getMatch("totalRenameNGtitle");
			String[] r = EasyRewriter.getReplace("totalRenameHelperConf");
			String r0 = ((r == null || r.length < 1) ? "IGNORE: L[[h \"$0\"" : r[0]);
			boolean found = false;
			JavaMatcher m0 = null;
			if (p1 != null && p1.length >= 1) {
				m0 = p1[0].matcher(title);				// BLACK MATCH
				found = m0.find();
				if (!found && p2 != null) {
					for (JavaPattern pat : p2) {
						m0 = pat.matcher(title);		// BLACK MATCH
						if (found = m0.find()) break;
					}
				}
			}
			if (p1 != null && p1.length >= 3) {
				JavaMatcher m1 = p1[1].matcher(title);	// WHITE MATCH
				JavaMatcher m2 = p1[2].matcher(smid);	// WHITE MATCH
				if (m1.find() || m2.find()) found = false;
			}
			if (found) {
				result = "RM ".concat(r0.replace("$0", m0.group()));
			} else {
				Matcher m = ExtUtil.matchNGtitle(title);
				if(m != null) result = "RM NGtitle: ".concat(m.group());
			}
		}
		// 폜Ȃ玩폜
		if (result.startsWith("RM") && Boolean.getBoolean("TR_cacheAutoRemove")) {
			if (Cache.remove(smid)) {
				util.info(smid.concat(" auto removed."));
			}
		}
		return result;
	}
	
	// t@C폜ǂݍ
	private static final Pattern DELETED_LINE_PATTERN = Pattern.compile(
			"^([a-z]{2}\\d+)($|[^\\d].*$)");
	public void loadDeleted(BufferedReader br) throws IOException {
		if (br != null) {
			Set<String> loaded = new java.util.HashSet<String>();
			String line;
			while ((line = br.readLine()) != null) {
				Matcher m = DELETED_LINE_PATTERN.matcher(line);
				if (m.find()) loaded.add(m.group(1));
			}
			deleted = Collections.synchronizedSet(loaded);
		} else {
			deleted = null;
		}
	}
	
	// t@C폜ǂݍ
	public void loadRemoveInfo(BufferedReader br) throws IOException {
		if (rminfo == null) {
			rminfo = Collections.synchronizedSortedMap(
					new java.util.TreeMap<String, String>(this));
		} else {
			rminfo.clear();
		}
		if (br != null) {
			String line;
			while ((line = br.readLine()) != null) {
				String data[] = line.split("\t", 2);
				rminfo.put(data[0], (data.length < 2) ? "(NULL)" : data[1]);
				checkRemove(data[0], null); // 폜
			}
		}
	}
	
	// 폜̕Ԃ
	private String rminfoString() {
		StringBuilder sb = new StringBuilder(1024*10);
		for (Map.Entry<String,String> e : rminfo.entrySet()) {
			sb.append(e.getKey());
			if (e.getValue() != null) {
				sb.append("\t").append(e.getValue());
			}
			sb.append("\r\n");
		}
		return sb.toString();
	}
	
	// 폜̎擾
	private String getRemoveInfo(String smid) {
		return rminfo.get(stripLow(smid));
	}
	
	// 폜̐ݒ
	private void setRemoveInfo(String smid, String reason) {
		if (reason != null) {
			rminfo.put(stripLow(smid), reason);
		} else {
			rminfo.remove(stripLow(smid));
		}
		rminfoSaveNeeded = true;
	}
	
	// IDGRm~[Ȃ疖low
	private String stripLow(String smid) {
		if (smid.endsWith("low"))
			return smid.substring(0, smid.length() - 3);
		return smid;
	}
	
	// cacheCommentExtensionRgLbV폜
	private void removeCommentCache(String smid) throws IOException {
		File commentFolder = ExtUtil.cacheFolderFile("#comment");
		if (!commentFolder.exists()) return;
		File[] commentCaches = commentFolder.listFiles();
		for (int i = 0; i < commentCaches.length; i++) {
			File f = commentCaches[i];
			String fn = f.getName();
			if (f.isFile() && fn.startsWith(smid.concat("_")) && fn.endsWith(".xml")) {
				if(f.delete()) util.info("Remove Comment: ".concat(fn));
			}
		}
	}
	
	// 폜Ɋ܂܂邩H
	private boolean containsDeleted(String smid) {
		util.loadFile(deletedFileInfo, "loadDeleted", FILE_CHARSET);
		if (deleted != null && deleted.contains(smid)) {
			return true;
		}
		// LbṼpXdeleted܂܂Ăꍇ폜Ȃ
		File f = Cache.getId2File().get(smid);
		if (f != null && f.getPath().contains("deleted")) {
			return true;
		}
		return false;
	}
	
}
